
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include <ctype.h>
#include "xstd/h/iostream.h"
#include "xstd/h/iomanip.h"

#include "base/ObjId.h"
#include "base/RndPermut.h"
#include "base/polyLogCats.h"
#include "base/polyLogTags.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/Viserv.h"
#include "runtime/PubWorld.h"
#include "runtime/StatPhase.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/LogComment.h"
#include "runtime/ErrorMgr.h"
#include "runtime/EphPortMgr.h"
#include "runtime/ExpPortMgr.h"
#include "runtime/PopModel.h"
#include "runtime/polyBcastChannels.h"
#include "runtime/polyErrors.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/globals.h"
#include "csm/ContentSel.h"
#include "csm/ContentCfg.h"
#include "pgl/RobotSym.h"
#include "pgl/PglSemx.h"
#include "dns/DnsMgr.h"
#include "client/CltConnMgr.h"
#include "client/CltOpts.h"
#include "client/CltXact.h"
#include "client/IcpCltXact.h"
#include "client/ServerRep.h"
#include "client/PrivCache.h"
#include "client/CltCfg.h"
#include "client/cltReqTypes.h"
#include "client/Client.h"


class IcpAgentCfg;

struct SharedCfgItem { const RobotSym *sym; CltCfg *cfg; };
static Array<SharedCfgItem> SharedCfgs;

XactFarm<CltXact> *Client::TheXacts = 0;
ObjFarm<IcpCltXact> Client::TheIcpXacts;

CltSharedCfgs *Client::TheSharedCfgs = new CltSharedCfgs;
Array<PortMgr*> Client::ThePortMgrs;

int Client::TheClientCnt = 0;


Client::Client(): thePrivCache(0), theCfg(0), 
	theConnMgr(0), theDnsMgr(0), theIcpClient(0),
	theCcXactLvl(0), theLaunchCount(0), theExtraLaunchLvl(0) {

	theChannels.append(TheInfoChannel);
	theChannels.append(TheLogCfgChannel);
	theChannels.append(TheLogStateChannel);
	startListen();
}

Client::~Client() {
	delete theConnMgr;
	delete theDnsMgr;
	delete thePrivCache;
}

void Client::configure(const RobotSym *cfg, const NetAddr &aHost) {
	Assert(TheXacts);

	SockOpt opt;
	Agent::configure(cfg, aHost, opt);
	theHost.port(-1); // clients do not care about port number

	theCfg = TheSharedCfgs->getConfig(cfg);

	thePrivWorlds.stretch(theCfg->originCount());
	for (int i = 0; i < theCfg->originCount(); ++i)
		thePrivWorlds.append(PrivWorld(theId));

	int pcCap = 0;
	if (cfg->privCache(pcCap))
		thePrivCache = new PrivCache(pcCap);

	theConnMgr = new CltConnMgr;
	theConnMgr->configure(opt, cfg,
		TheCltOpts.thePxyAddr ? 1 : theCfg->originCount());
	theConnMgr->portMgr(getPortMgr(true));
	theConnMgr->idleTimeout(cfg->idlePconnTimeout());

	theDnsMgr = new DnsMgr(this);
	theDnsMgr->configure(cfg->dnsResolver());

	if (theCfg->thePeerHttp && !theCfg->thePeerIcp) {
		cerr << cfg->loc() << "the HTTP peer is at " << theCfg->thePeerHttp << ", but where is the ICP peer?" << endl;
		exit(-3);
	}

	TheXacts->limit(1024); // magic, no good way to estimate
}

void Client::icpClient(IcpClient *anIcpClient) {
	Assert(!theIcpClient);
	theIcpClient = anIcpClient;
}

void Client::start() {
	Assert(!theLaunchCount);
	Assert(theConnMgr);
	Assert(theDnsMgr);

	Agent::start();

	// first client that gets here will lock the phase
	if (!TheClientCnt)
		TheStatPhaseMgr->lock();
	TheClientCnt++;

	theDnsMgr->start();

	theCfg->startWarmup(); // warmup plan created if needed

	if (theCfg->theLaunchWin > 0) {
		// uniformly distributed start delay
		static RndGen rng;
		const double delay = rng(0.0, theCfg->theLaunchWin.secd());
		sleepFor(Time::Secd(delay));
	} else {
		scheduleLaunch(TheClock);
	}
}

void Client::noteInfoEvent(BcastChannel *ch, InfoEvent ev) {
	Assert(ch == TheInfoChannel);
	if (ev == ieWssFreeze) {
		for (int i = 0; i < thePrivWorlds.count(); ++i)
			thePrivWorlds[i].freezeWss();

		static bool frozen = false;
		if (!frozen) {
			FreezeSrvWss();
			frozen = true;
			const BigSize meanFillSz = TheStatPhaseMgr->fillCnt() ?
				TheStatPhaseMgr->fillSz()/TheStatPhaseMgr->fillCnt() :
				BigSize(0);
			const int pubWss = PubWorld::PositiveWss();
			const int privWss = PrivWorld::TheWss;
			Comment(5) << "fyi: min `direct' objects in working set:"
				<< " global public: "
					<< pubWss << " (" << (meanFillSz*pubWss) << ')'
				<< " local private: "
					<< privWss << " (" << (meanFillSz*privWss) << ')'
				<< endc;
		}
	}
}

void Client::noteLogEvent(BcastChannel *ch, OLog &log) {
	if (ch == TheLogCfgChannel) {
		log << bege(lgCltCfg, lgcCltSide);
		// XXX: implement (CltCfgRec)
		log << ende;
	} else
	if (ch == TheLogStateChannel) {
		log << bege(lgCltState, lgcCltSide)
			<< theSeqvId 
			<< theLaunchCount
			<< thePrivWorlds
			<< ende;
	}
}

void Client::noteXactDone(CltXact *x) {
	Assert(x);

	const ObjId oid(x->oid());
	const bool retry = x->needRetry();

	theConnMgr->put(x->conn());
	TheXacts->put(x);
	theCcXactLvl--;
	Assert(theCcXactLvl >= 0);

	// retry immediately if needed and possible
	// XXX: request type may be different this time
	if (retry && tryLaunch(genXact(oid)))
		return;
		
	// push waiting xactions forward
	if (theLaunchDebts.count() && !theConnMgr->atHardConnLimit())
		resumeXact();
}

CltXact *Client::fetch(const ObjId &oid, DistrPoint *dp) {
	Assert(oid);
	CltXact *x = genXact(oid);
	x->cacheDistrPoint(dp);
	if (tryLaunch(x))
		return x;
	return 0;
}

// must not be called directly but rather through tryLaunch
bool Client::launch(CltXact *x) {
	Assert(x);
	Assert(!theConnMgr->atHardConnLimit());

	const NetAddr &connAddr = x->nextHop();
	Assert(connAddr);
	Assert(isdigit(connAddr.addrA()[0]));

	if (Connection *conn = theConnMgr->get(connAddr)) {
		theLaunchCount++;
		theCcXactLvl++;
		x->exec(this, conn);
		return true;
	}

	if (ReportError(Error::Last()))
		Comment(1) << theHost << " failed to connect to " << connAddr << endc;

	launchFailed(x);
	return false;
}

// will launch the xaction (if possible) or postpone it (if not)
// returns true if the xaction will be launched [eventually]
bool Client::tryLaunch(CltXact *x) {
	Assert(x);

	if (!x->nextHop()) {

		// should we ask peers for the best server?
		// XXX: remove askedPeer; the x->nextHop() should be enough?
		if (theCfg->thePeerIcp && !x->askedPeer()) {
			askPeer(theCfg->thePeerIcp, x);
			return true;
		}

		// next hop address should be known at this time 
		// (but my not be resolved yet)
		x->nextHop(nextHopAddr(x));

		// should we lookup next hop address?
		if (theDnsMgr->needsLookup(x->nextHop())) {
			lookupAddr(x);
			return true;
		}
	}

	// check if we should postpone the xaction
	if (theConnMgr->atHardConnLimit()) {
		if (theCfg->theWaitXactLmt < 0 || theLaunchDebts.count() < theCfg->theWaitXactLmt) {
			suspendXact(x);
			return true;
		}

		if (ReportError(errTooManyWaitXact))
			Comment(3) << "xactions active: " << theCcXactLvl 
				<< " waiting: " << theLaunchDebts.count() 
				<< " limit: " << theCfg->theWaitXactLmt << endc;
		launchFailed(x);
		return false;
	}

	return launch(x);
}

void Client::suspendXact(CltXact *x) {
	theLaunchDebts.append(x);
	Broadcast(TheWaitBegChannel, x);
}

void Client::resumeXact() {
	CltXact *x = theLaunchDebts.pop();
	Broadcast(TheWaitEndChannel, x);
	launch(x);
}

void Client::launchFailed(CltXact *x) {
	Assert(x);
	Broadcast(TheXactErrChannel, x);
	x->noteAbort();
	TheXacts->put(x);
}

CltXact *Client::genXact(const ObjId &oid) {
	CltXact *x = TheXacts->get();
	x->oid(oid);
	return x;
}

CltXact *Client::genXact() {
	ObjId oid;
	genOid(oid);
	return genXact(oid);
}

void Client::genOid(ObjId &oid) {
	selectViserv(oid);
	selectTarget(oid);
	selectObj(oid);
	selectContType(oid);
	selectReqType(oid);

	if (theCfg->genUniqUrls)
		oid.world(UniqId::Create()); // changes every time
}

void Client::selectViserv(ObjId &oid) {
	oid.viserv(theCfg->selectViserv());
	Assert(oid.viserv() >= 0);
}

void Client::selectTarget(ObjId &oid) {
	const int visName = TheViservs.nameIdxAt(oid.viserv());
	const NetAddr &addr = TheAddrMap->selectAddr(visName);

	int idx = -1;
	Assert(TheHostMap->find(addr, idx));
	oid.target(idx);

	const HostCfg *hcfg = TheHostMap->at(oid.target());
	Assert(hcfg->theServerRep);
}
	

void Client::selectObj(ObjId &oid) {
	static RndGen rng;

	OidGenStat &oidGenStat = TheStatPhaseMgr->oidGenStat();

	const bool needPub = rng.event(theCfg->thePublicRatio);
	const bool needRepeat = rng.event(theCfg->theRecurRatio*
		TheStatPhaseMgr->recurFactor().current(TheClock, theLaunchCount));
	oidGenStat.recordNeed(needRepeat, needPub);

	PrivWorld &privWorld = thePrivWorlds[oid.viserv()];
	PubWorld &pubWorld = *ThePubWorlds[oid.viserv()];

	const bool privCanRep = privWorld.canRepeat();
	const bool privCanProd = privWorld.canProduce();

	const bool pubCanRep = pubWorld.canRepeat();
	const bool pubCanProd = pubWorld.canProduce();

	const bool canRep = privCanRep || pubCanRep;
	const bool canProd = privCanProd || pubCanProd;

	// the logic gives priority to repeatOid goal rather than to genPubOid
	if (canRep && (needRepeat || !canProd)) {
		if (pubCanRep && (needPub || !privCanRep)) {
			pubWorld.repeat(oid, theCfg->thePopModel, rng);
			oidGenStat.recordGen(true, true);
			return;
		}

		Assert(privCanRep); 
		privWorld.repeat(oid, theCfg->thePopModel, rng);
		oidGenStat.recordGen(true, false);
		return;
	}

	// new public
	if (pubCanProd && (needPub || !privCanProd)) {
		pubWorld.produce(oid, rng);
		oidGenStat.recordGen(false, true);
		return;
	}

	// new private object (last resort, never fails)
	privWorld.produce(oid, rng);
	oidGenStat.recordGen(false, false);
}

void Client::selectContType(ObjId &oid) {
	const HostCfg *hcfg = TheHostMap->at(oid.target());
	Assert(hcfg);
	Assert(hcfg->theContent);
	const ContentCfg *ccfg = hcfg->theContent->getDir(oid);
	oid.type(ccfg->id());
}

void Client::selectReqType(ObjId &oid) {
	static RndGen rng;
	const TransFactor &tf = TheStatPhaseMgr->specialMsgFactor();
	const int reqType = rng.event(tf.current(TheClock, theLaunchCount)) ?
		(int)theCfg->theReqTypeSel->trial() : rqtBasic;
	oid.ims(reqType == rqtIms);
	oid.reload(reqType == rqtReload);
	oid.rediredReq(false);
}

void Client::noteRedirect(CltXact *, const ObjId &oid) {
	Assert(oid.rediredReq());
	if (tryLaunch(genXact(oid)))
		theExtraLaunchLvl++;
}

void Client::noteEmbedded(CltXact *, const ObjId &oid) {
	static RndGen rng;
	if (!rng.event(theCfg->theEmbedRecurRatio))
		return;

	if (tryLaunch(genXact(oid)))
		theExtraLaunchLvl++;
}

void Client::askPeer(const NetAddr &addr, CltXact *x) {
	IcpCltXact *q = TheIcpXacts.get();
	q->reason(x);
	q->exec(this, addr);
}

void Client::notePeerAsked(IcpCltXact *q) {
	Assert(q);

	CltXact *x = q->reason(this);
	x->usePeer(q->hit());
	/* pass ICP stats here ... */

	TheIcpXacts.put(q);

	tryLaunch(x);
}

void Client::lookupAddr(CltXact *x) {
	theDnsMgr->lookup(x->nextHop(), x);
}

void Client::noteAddrLookup(const NetAddr &addr, CltXact *x) {
	Assert(x);

	// successful lookup?
	if (addr) {
		x->nextHop(addr); // update
		tryLaunch(x);
	} else {
		launchFailed(x);
	}
}

// select the address for the next hop connection
const NetAddr &Client::nextHopAddr(const CltXact *x) const {

	if (x->usePeer() && theCfg->thePeerHttp)
		return theCfg->thePeerHttp;

	if (TheCltOpts.thePxyAddr)
		return TheCltOpts.thePxyAddr;

	return TheViservs.nameAt(x->oid().viserv());
}

PortMgr *Client::getPortMgr(bool bind2iface) {
	const NetAddr addr = bind2iface ? theHost : NetAddr();

	// check if we already have a port mgr for host
	for (int i = 0; i < ThePortMgrs.count(); ++i) {
		if (ThePortMgrs[i]->addr() == addr)
			return ThePortMgrs[i];
	}

	// create new port manager
	PortMgr *mgr = TheCltOpts.thePorts.set() ?
		(PortMgr*) new ExpPortMgr(addr, TheCltOpts.thePorts.lo(), TheCltOpts.thePorts.hi()) :
		(PortMgr*) new EphPortMgr(addr);
	ThePortMgrs.append(mgr);
	return mgr;
}


int Client::logCat() const {
	return lgcCltSide;
}

void Client::FreezeSrvWss() {
	for (int i = 0; i < ThePubWorlds.count(); ++i)
		ThePubWorlds[i]->freezeWss();
}

void Client::Farm(XactFarm<CltXact> *aFarm) {
	Assert(!TheXacts && aFarm);
	TheXacts = aFarm;
}

void Client::LogState(OLog &log) {
	log << bege(lgSrvRepState, lgcCltSide);
	OLogStorePtrs(log, ThePubWorlds);
	log << ende;
}
