
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "xstd/h/iomanip.h"

#include "xstd/rndDistrs.h"
#include "base/RndPermut.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/Viserv.h"
#include "runtime/PubWorld.h"
#include "runtime/PopModel.h"
#include "runtime/LogComment.h"
#include "pgl/RobotSym.h"
#include "client/WarmupPlan.h"
#include "client/ServerRep.h"
#include "client/CltCfg.h"
#include "client/cltReqTypes.h"


CltCfg::CltCfg(): theRobot(0), theOriginSel(0), theReqTypeSel(0), thePopModel(0),
	theWarmupPlan(0),
	thePublicRatio(-1), theRecurRatio(-1), theEmbedRecurRatio(-1),
	theWaitXactLmt(-1), theIcpPort(-1), genUniqUrls(false), didWarmup(false) {
}

void CltCfg::configure(const RobotSym *aRobot) {
	Assert(!theRobot && aRobot);
	theRobot = aRobot;

	theRobot->waitXactLimit(theWaitXactLmt);
	thePeerHttp = theRobot->peerHttp();
	thePeerIcp = theRobot->peerIcp();
	theRobot->icpPort(theIcpPort);

	theLaunchWin = theRobot->launchWin();
	theRobot->publicRatio(thePublicRatio);
	theRobot->recurRatio(theRecurRatio);
	theRobot->embedRecurRatio(theEmbedRecurRatio);
	theRobot->uniqueUrls(genUniqUrls);

	if (PopModelSym *pms = theRobot->popModel()) {
		thePopModel = new PopModel;
		thePopModel->configure(pms);
	}

	configureRTypes();
	configureOrigins();

	if (!thePopModel && theRecurRatio > 0)
		cerr << theRobot->loc() << "popularity model must be specified for"
			<< " positive recurrence ratio (robot " << theRobot->kind() << ')'
			<< endl << xexit;
}

void CltCfg::configureRTypes() {
	theReqTypeSel = theRobot->reqTypes(CltReqTypeIdx());
	if (!theReqTypeSel)
		theReqTypeSel = new ConstDistr(new RndGen, rqtBasic); // default
	theReqTypeSel->rndGen(GenRndGen("client_req_types"));
}

void CltCfg::configureOrigins() {
	Array<NetAddr*> origNames;

	RndDistr *iad = 0;
	if (!theRobot->origins(origNames, theOriginSel) || !origNames.count()) {
		if (!theRobot->reqInterArrival(iad) || iad)
			cerr << theRobot->loc() << "no origin addresses specified for active robot " 
				<< theRobot->kind() << endl << xexit;
		return;
	}

	Assert(theOriginSel);
	theOriginSel->rndGen(GenRndGen("client_origins"));

	theOrigins.stretch(origNames.count());
	for (int i = 0; i < origNames.count(); ++i) {
		addOrigName(*origNames[i]);
		cAddServers(theOrigins[i]);
	}
}

// converts origin name into viserv idx
void CltCfg::addOrigName(const NetAddr &oname) {
	int nameIdx = -1;
	if (!TheAddrMap->find(oname, nameIdx)) {
		if (oname.isDomainName())
			cerr << here << "visible server name " << oname
				<< " is not found in address maps" << endl << xexit;
		TheAddrMap->add(oname);
	}

	int viserv = -1;
	if (!TheViservs.find(nameIdx, viserv)) {
		TheViservs.add(nameIdx);
		ThePubWorlds.append(new PubWorld(UniqId::Create()));
	}

	theOrigins.append(viserv);
}

// creates new server representatives for vname
void CltCfg::cAddServers(int viserv) {
	const int visName = TheViservs.nameIdxAt(viserv);
	for (AddrMapAddrIter i = TheAddrMap->addrIter(visName); i; ++i) {
		const NetAddr &addr = i.addr();
		int idx = -1;
		if (!TheHostMap->find(addr, idx)) {
			Comment << theRobot->loc() << "Robot cannot find"
				<< " configuration for server@" << addr
				<< " (visible as " << i.name() << ")"
				<< endc << xexit;
		}

		// create representative if needed
		HostCfg *cfg = TheHostMap->at(idx);
		if (!cfg->theServerRep)
			cfg->theServerRep = new ServerRep(addr, idx);
	}
}

int CltCfg::selectViserv() {
	if (theWarmupPlan) {
		const int viservIdx = theWarmupPlan->selectViserv();
		if (viservIdx >= 0)
			return viservIdx;
		stopWarmup();
	}

	const int viservIdx = (int)theOriginSel->trial();
	Assert(0 <= viservIdx && viservIdx <= theOrigins.count());
	return theOrigins[viservIdx];
}


// may be called multiple times
void CltCfg::startWarmup() {
	if (!theWarmupPlan && !didWarmup)
		theWarmupPlan = new WarmupPlan(theOrigins);
}

void CltCfg::stopWarmup() {
	Assert(theWarmupPlan && !didWarmup);
	delete theWarmupPlan;
	theWarmupPlan = 0;
	didWarmup = true;
}

/* CltSharedCfgs */

CltCfg *CltSharedCfgs::getConfig(const RobotSym *rs) {
	for (int i = 0; i < count(); ++i) {
		if (item(i)->theRobot == rs)
			return item(i);
	}
	return addConfig(rs);
}

CltCfg *CltSharedCfgs::addConfig(const RobotSym *rs) {
	CltCfg *cfg = new CltCfg;
	cfg->configure(rs);
	append(cfg);
	return cfg;
}

