
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "xstd/h/iomanip.h"

#include "pgl/AddrMapSym.h"
#include "runtime/AddrMapItems.h"
#include "runtime/AddrMap.h"


AddrMap *TheAddrMap = 0;


AddrMap::~AddrMap() {
	while (theMap.count()) delete theMap.pop();
}

void AddrMap::configure(const Array<AddrMapSym*> &maps) {
	for (int i = 0; i < maps.count(); ++i) {
		AddrMapSym &ms = *maps[i];
		Addrs names;
		Addrs addrs;

		ms.names(names);
		ms.addresses(addrs);

		const int nameCount = names.count();
		const int addrCount = addrs.count();

		// check that all addresses are IPs
		for (int a = 0; a < addrCount; ++a) {
			if (addrs[a]->isDomainName())
				cerr << ms.loc() << "the `address' part of an AddrMap "
					<< " must not contain domain names, got: " << *addrs[a]
					<< endl << xexit;
		}

		theMap.stretch(theMap.count() + nameCount);

		if (nameCount == 1 && addrCount == 1)
			config1to1(*names[0], *addrs[0]);
		else
		if (nameCount == 1 && addrCount > 1)
			configure1toN(*names[0], addrs);
		else
		if (nameCount > 1 && nameCount == addrCount)
			configNtoN(names, addrs);
		else
			cerr << ms.loc() << "only 1:1 and 1:N mappings are"
				<< " supported for address maps"
				<< " got " << nameCount << " names"
				<< " and " << addrCount << " addresses"
				<< endl << xexit;

		while (names.count()) delete names.pop();
		while (addrs.count()) delete addrs.pop();
	}
}

void AddrMap::config1to1(const NetAddr &name, const NetAddr &addr) {
	theMap.append(new Name2AddrMapItem(name, addr));
}

void AddrMap::configNtoN(const Addrs &names, const Addrs &addrs) {
	for (int i = 0; i < names.count(); ++i)
		config1to1(*names[i], *addrs[i]);
}

void AddrMap::configure1toN(const NetAddr &name, const Addrs &addrs) {
	theMap.append(new Name2AddrsMapItem(name, addrs));
}

const AddrMapItem &AddrMap::itemAt(int nameIdx) const {
	Assert(0 <= nameIdx && nameIdx < theMap.count());
	return *theMap[nameIdx];
}

int AddrMap::nameCount() const {
	return theMap.count();
}

const NetAddr &AddrMap::nameAt(int nameIdx) const {
	return itemAt(nameIdx).name();
}

const NetAddr &AddrMap::addrAt(int nameIdx, int addrIdx) const {
	const AddrMapItem &mi = itemAt(nameIdx);
	Assert(0 <= addrIdx && addrIdx < mi.addrCount());
	return mi.addrAt(addrIdx);
}

int AddrMap::addrCountAt(int nameIdx) const {
	return itemAt(nameIdx).addrCount();
}

bool AddrMap::find(const NetAddr &name, int &idx) const {
	// XXX: optimize the search!
	for (idx = 0; idx < theMap.count(); ++idx) {
		if (name == theMap[idx]->name())
			return true;
	}
	return false;
}

void AddrMap::add(const NetAddr &name) {
	if (name.isDomainName()) {
		cerr << here << "domain name `" << name 
			<< "' used where an IP address was expected"
			<< endl << xexit;
	}
	theMap.append(new Name2NameMapItem(name));
}

bool AddrMap::findAddr(const NetAddr &addr) const {
	for (int idx = 0; idx < theMap.count(); ++idx) {
		for (AddrMapAddrIter i = addrIter(idx); i; ++i) {
			if (i.addr() == addr)
				return true;
		}
	}
	return false;
}

const NetAddr &AddrMap::selectAddr(int nameIdx) const {
	return itemAt(nameIdx).selectAddr();
}

AddrMapAddrIter AddrMap::addrIter(int nameIdx) const {
	Assert(0 <= nameIdx && nameIdx < theMap.count());
	return AddrMapAddrIter(*this, nameIdx);
}


/* AddrMapAddrIter */

AddrMapAddrIter::AddrMapAddrIter(const AddrMap &aMap, int aName):
	theMap(aMap), theName(aName), theAddr(0) {
}

const NetAddr &AddrMapAddrIter::addr() {
	return theMap.addrAt(theName, theAddr);
}

const NetAddr &AddrMapAddrIter::name() {
	return theMap.nameAt(theName);
}

bool AddrMapAddrIter::atEnd() const {
	return theAddr >= theMap.addrCountAt(theName);
}
