
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "base/polygraph.h"

#include "base/OLog.h"
#include "xstd/String.h"
#include "xstd/Clock.h"
#include "xstd/gadgets.h"
#include "runtime/StrIdentifier.h"
#include "runtime/HttpDate.h"
#include "runtime/httpText.h"
#include "runtime/httpHdrs.h"


/* internal type to store static parsing info */
class MsgHdrParsTab {
	public:
		MsgHdrParsTab();
		~MsgHdrParsTab();

	public:
		StrIdentifier *ids;
		Array<MsgHdr::Parser*> *parsers;
};


MsgHdrParsTab *ReqHdr::TheParsTab = 0;
MsgHdrParsTab *RepHdr::TheParsTab = 0;


/* MsgHdr */

MsgHdr::MsgHdr(const MsgHdrParsTab &aTab): theParsTab(aTab) {
	reset();
}

void MsgHdr::reset() {
	theHdrSize = 0;
	theContSize = -1;
	theDate = Time();
	theGroupId.clear();
	theXactId.clear();
	theTarget = NetAddr();
	theRemWorld.reset();
	thePhaseSyncPos = -1;
	theHttpVer = -1;
	theKeepAliveConn = kaDefault;
	isCachable = true;

	theBufBeg = theBufEnd = theSrchPtr = 0;
	theFields.reset();
	theSrchState = ssFirst;

	// do not reset parsing tables
}

bool MsgHdr::keepAlive() const {
	if (theHttpVer >= 1.1) // 1.1: keep unless told otherwise
		return theKeepAliveConn != kaNo;
	else                   // 1.0: keep if explicitly told so
		return theKeepAliveConn == kaYes;
}

// note: buf does not have to be zero-terminated!
bool MsgHdr::parse(const char *buf, Size sz) {
	if (!theBufBeg) { // have not started the search yet
		theBufBeg = theSrchPtr = buf;
		Assert(theSrchState == ssFirst);
	} else {           // continue search
		Assert(theBufBeg == buf);
		Assert(theSrchState != ssFound);
	}

	theBufEnd = buf + sz; // to be refined later
	do {
		// search for LF
		if (theSrchState == ssFirst) {
			do {
				if (*theSrchPtr++ == '\n') {
					theSrchState = ssSkip;
					break;
				}
			} while (theSrchPtr < theBufEnd);
		}

		// LF after skipping optional CRs means and-of-headers
		while (theSrchState == ssSkip && theSrchPtr < theBufEnd) {
			if (*theSrchPtr == '\n') {
				theSrchState = ssFound;
			} else
			if (*theSrchPtr != '\r') {
				theFields.append(theSrchPtr); // start of a header!
				theSrchState = ssFirst;
			}
			++theSrchPtr;
		}
	} while (theSrchPtr < theBufEnd && theSrchState != ssFound);

	if (theSrchState != ssFound)
		return false;

	// found end-of-headers!
	theBufEnd = theSrchPtr;
	theHdrSize = theBufEnd - theBufBeg;

	// now parse known fields
	// luckily, we already know field starts!
	parseFields();
	return true;
}

void MsgHdr::parseFields() {
	const char *eoh = theBufEnd;
	parseRLine(theBufBeg, theFields.count() ? theFields[0] : theBufEnd);

	for (int i = theFields.count()-1; i >= 0; --i) {
		const char *hdr = theFields[i];
		const int len = eoh-hdr; // approximate (includes crlfs)
		const int id = theParsTab.ids->lookup(hdr, len);
		if (id > 0) {
			const char *val = hdr + theParsTab.ids->string(id).len();
			while (isspace(*val)) ++val;
			Parser p = *theParsTab.parsers->item(id);
			(this->*p)(val, eoh);
		}
		eoh = hdr;
	}
}

bool MsgHdr::parseUniqId(const char *&buf, const char *, UniqId &id) {
	int a = 0, b = 0, c = 0;
	const char *p = 0;
	if (isInt(buf, a, &p, 16) && *p == '.' &&
		isInt(p+1, b, &p, 16) && *p == ':' &&
		isInt(p+1, c, &buf, 16))
		return (id = UniqId(a, b, c)) != 0;

	return false;
}

bool MsgHdr::parseWorld(const char *buf, const char *eoh, ObjWorld &world) {
	UniqId id;
	if (!parseUniqId(buf, eoh, id) || *buf != ' ')
		return false;

	int size = -1, wss = -1;
	int hotPos = -1;
	const char *p = buf;

	if (isInt(p+1, wss, &p) && *p == '/' &&
	    isInt(p+1, size, &p) && *p == ' ' &&
	    isInt(p+1, hotPos)) {
		world.id(id);
		world.size(size);
		world.wss(wss);
		world.hotPos(hotPos);
		return true;
	}

	return false;
}

bool MsgHdr::parseNetAddr(const char *buf, const char *eoh, NetAddr &addr) {
	int port = 80; // XXX: default should depend on protocol
	if (const char *p = StrBoundChr(buf, ':', eoh)) {
		if (isInt(p + 1, port))
			eoh = p;
		else
			return false; // malformed port number
	} else {
		// remove CRLF (if :port is present that is done automagically)
		while (buf < eoh && eoh[-1] == '\n') --eoh;
		while (buf < eoh && eoh[-1] == '\r') --eoh;
		if (buf == eoh)
			return false;
	}

	addr.port(port);
	return isdigit(*buf) ?
		parseIp(buf, eoh, addr) : parseDname(buf, eoh, addr);
}

bool MsgHdr::parseDname(const char *buf, const char *eoh, NetAddr &addr) {
	if (buf < eoh) {
		String name;
		name.append(buf, eoh-buf);
		addr.addr(name);
		return true;
	}

	return false;
}

bool MsgHdr::parseIp(const char *buf, const char *eoh, NetAddr &addr) {
	// convert a.b.c.d to an int address
	// this is a poor man's inet_aton()...
	int acc = 0;
	if (!isInt(buf, acc, &buf))
		return false;
	for (int i = 2; i >= 0 && buf < eoh; --i) {
		if (*buf != '.')
			return false;
		int octet;
		if (!isInt(buf+1, octet, &buf))
			return false;
		if (octet < 0 || octet > 255)
			return false;
		acc <<= 8;
		acc |= octet;
	}

	if (buf != eoh)
		return false; // trailing garbage

	in_addr name;
	name.s_addr = htonl(acc);
	addr.addr(name);
	return true;
}

bool MsgHdr::parseHostInUri(const char *&start, const char *eorl, NetAddr &host) {
	const char *buf = start;
	if (const char *n = StrBoundChr(buf, ':', eorl)) {
		if (*++n == '/' && *++n == '/') {
			buf = n+1;

			const char *end = StrBoundChr(buf, '/', eorl);
			if (!end)
				end = eorl;

			if (parseNetAddr(buf, end, host)) {
				start = end;
				return true;
			}
		}
	}
	return false;
}

bool MsgHdr::parseUri(const char *&buf, const char *end, HttpUri &uri) {
	// see if there is a protocol://host prefix
	if (*buf != '/')
		parseHostInUri(buf, end, uri.host);

	// user-specified url_prefix may contain world-id tag;
	// search for the first tag that is followed by a valid wid
	while (const char *wid = StrBoundChr(buf, 'w', end)) {
		if (parseUniqId(buf = wid+1, end, uri.oid.world()))
			break;
	}

	if (const char *tid = StrBoundChr(buf, 't', end))
		isInt(tid + 1, uri.oid.type(), &buf, 16);
	if (const char *oid = StrBoundChr(buf, '_', end))
		isInt(oid + 1, uri.oid.name(), &buf, 16);
	return true;
}

bool MsgHdr::parseContLen(const char *buf, const char *) {
	theContSize = xatoi(buf, -1);
	return theContSize >= 0;
}

bool MsgHdr::parseDate(const char *buf, const char *eoh) {
    theDate = HttpDateParse(buf, eoh - buf);
    return theDate >= 0;
}

bool MsgHdr::parsePragma(const char *buf, const char *) {
	if (!strncasecmp("no-cache", buf, 8))
		isCachable = false;
	else
		return false;
	return true;
}

bool MsgHdr::parseCControl(const char *buf, const char *) {
	if (!strncasecmp("no-cache", buf, 8)) {
		isCachable = false;
		return true;
	}
	return false;
}

bool MsgHdr::parseXXact(const char *buf, const char *eoh) {
	return
		parseUniqId(buf, eoh, theGroupId) && *buf == ' ' &&
		parseUniqId(++buf, eoh, theXactId);
}

bool MsgHdr::parseXRemWorld(const char *buf, const char *eoh) {
	return parseWorld(buf, eoh, theRemWorld);
}

bool MsgHdr::parseXPhaseSyncPos(const char *buf, const char *) {
	thePhaseSyncPos = xatoi(buf, 0);
	return true;
}

bool MsgHdr::parseXTarget(const char *buf, const char *eoh) {
	return parseNetAddr(buf, eoh, theTarget);
}

/* XXX: Connection and other headers may have a _list_ of options */

bool MsgHdr::parseConnection(const char *buf, const char *) {
	if (!strncasecmp("close", buf, 5))
		theKeepAliveConn = kaNo;
	else
	if (!strncasecmp("keep", buf, 4))
		theKeepAliveConn = kaYes;
	else
		return false;
	return true;
}

// adds definitions common to replies and requests
void MsgHdr::Configure(MsgHdrParsTab &tab) {
	AddParser(hfpDate, &MsgHdr::parseDate, tab);
	AddParser(hfpContLength, &MsgHdr::parseContLen, tab);
	AddParser("Cache-Control:", &MsgHdr::parseCControl, tab);
	AddParser("Connection:", &MsgHdr::parseConnection, tab);
	AddParser("Pragma:", &MsgHdr::parsePragma, tab);
	AddParser("Proxy-Connection:", &MsgHdr::parseConnection, tab);
	AddParser(hfpXXact, &MsgHdr::parseXXact, tab);
	AddParser(hfpXRemWorld, &MsgHdr::parseXRemWorld, tab);
	AddParser(hfpXPhaseSyncPos, &MsgHdr::parseXPhaseSyncPos, tab);
	AddParser(hfpXTarget, &MsgHdr::parseXTarget, tab);
}

int MsgHdr::AddParser(const String &field, Parser parser, MsgHdrParsTab &where) {
	Assert(field.len());
	const String trimmedField = isspace(field.last()) ? field(0, field.len()-1) : field;
	Assert(trimmedField.len());

	const int id = where.ids->add(trimmedField);
	where.parsers->put(new Parser(parser), id);
	return id;
}

void MsgHdr::store(OLog &log) const {
	log
		<< theHdrSize
		<< (int)theDate.sec()
		<< theContSize
		<< theGroupId
		<< theXactId
		<< theTarget
		<< theHttpVer
		<< (int)theKeepAliveConn
		<< isCachable
		;
}

// these should never be called
inline bool dontCallMe() { Assert(0); return false; }
bool MsgHdr::parseHost(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseLocation(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseLMT(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseExpires(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseIms(const char *, const char *) { return dontCallMe(); }
bool MsgHdr::parseXLocWorld(const char *, const char *) { return dontCallMe(); }


/* ReqHdr */

ReqHdr::ReqHdr(): MsgHdr(*TheParsTab), isHealthCheck(false) {
}

void ReqHdr::reset() {
	MsgHdr::reset();
	theUri = HttpUri();
	theIms = Time();
	theLocWorld.reset();
	isHealthCheck = false;
}

bool ReqHdr::parseRLine(const char *buf, const char *eorl) {
	if (strncasecmp("GET ", buf, 4) == 0) {
		buf += 4;

		// a "well-known" health check uri
		static const String health = "/health";
		isHealthCheck = health.casePrefixOf(buf, eorl-buf);

		parseUri(buf, eorl, theUri);
		if (const char *ver = StrBoundChr(buf, ' ', eorl))
			isDbl(ver + 6, theHttpVer);
		return true;
	}
	return false;
}

bool ReqHdr::parseHost(const char *buf, const char *eoh) {
	return parseNetAddr(buf, eoh, theUri.host);
}

bool ReqHdr::parseIms(const char *buf, const char *eoh) {
	theIms = HttpDateParse(buf, eoh - buf);
	return theIms >= 0;
}

bool ReqHdr::parseXLocWorld(const char *buf, const char *eoh) {
	return parseWorld(buf, eoh, theLocWorld);
}

void ReqHdr::store(OLog &log) const {
	MsgHdr::store(log);
	log << theUri.host << theUri.oid << (int)theIms.sec();
}

void ReqHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser(hfpHost, &MsgHdr::parseHost, *TheParsTab);
	AddParser(hfpIMS, &MsgHdr::parseIms, *TheParsTab);
	AddParser(hfpXLocWorld, &MsgHdr::parseXLocWorld, *TheParsTab);
}

void ReqHdr::Clean() {
	delete TheParsTab;
	TheParsTab = 0;
}


/* RepHdr */

RepHdr::RepHdr(): MsgHdr(*TheParsTab), theStatus(scUnknown) {
}

void RepHdr::reset() {
	MsgHdr::reset();
	theLocn = HttpUri();
	theLMT = theExpires = Time();
	theStatus = scUnknown;
}

/* RFC 2616: All responses to the HEAD request method MUST NOT include
 * a message-body. All 1xx , 204, and 304 responses MUST NOT include a
 * message-body. All other responses do include a message-body.
 */
bool RepHdr::expectBody() const {
	// note: we cannot handle the HEAD case here; the caller should
	if ((100 <= theStatus && theStatus < 200) ||
		theStatus == sc204_NoContent ||
		theStatus == sc304_NotModified)
		return false;
	return true;
}

Time RepHdr::calcLmt() const {
	if (theLMT >= 0)
		return theLMT;
	if (theDate >= 0)
		return theDate;
	return TheClock;
}

bool RepHdr::redirect() const {
	return 
		theStatus == sc300_Choices ||
		theStatus == sc302_Found ||
		theStatus == sc303_Other ||
		theStatus == sc307_Tmp_Redir;
}

bool RepHdr::parseRLine(const char *buf, const char *) {
	// what is the actual version format in the specs??
	if (strncasecmp("HTTP/", buf, 5) != 0)
		return false;
	if (!isDbl(buf+5, theHttpVer, &buf))
		return false;
	return isInt(buf+1, theStatus);
}

bool RepHdr::parseLocation(const char *buf, const char *eoh) {
	return parseUri(buf, eoh, theLocn);
}

bool RepHdr::parseLMT(const char *buf, const char *eoh) {
    theLMT = HttpDateParse(buf, eoh - buf);
    return theLMT >= 0;
}

bool RepHdr::parseExpires(const char *buf, const char *eoh) {
    theExpires = HttpDateParse(buf, eoh - buf);
    return theExpires >= 0;
}

void RepHdr::store(OLog &log) const {
	MsgHdr::store(log);
	log
		<< theStatus
		<< (int)theLMT.sec()
		<< (int)theExpires.sec()
		;
}

void RepHdr::Configure() {
	TheParsTab = new MsgHdrParsTab();
	MsgHdr::Configure(*TheParsTab);
	AddParser(hfpLocation, &MsgHdr::parseLocation, *TheParsTab);
	AddParser(hfpLmt, &MsgHdr::parseLMT, *TheParsTab);
	AddParser(hfpExpires, &MsgHdr::parseExpires, *TheParsTab);
}

void RepHdr::Clean() {
	delete TheParsTab;
	TheParsTab = 0;
}


/* MsgHdrParsTab */

MsgHdrParsTab::MsgHdrParsTab() {
	ids = new StrIdentifier;
	parsers = new Array<MsgHdr::Parser*>;
}

MsgHdrParsTab::~MsgHdrParsTab() {
	delete ids; ids = 0;
	while (parsers->count()) delete parsers->pop();
	delete parsers; parsers = 0;
}
